# Preferences

Utilities for storing WordPress preferences.

## Installation

Install the module

```bash
npm install @wordpress/preferences --save
```

_This package assumes that your code will run in an **ES2015+** environment. If you're using an environment that has limited or no support for such language features and APIs, you should include [the polyfill shipped in `@wordpress/babel-preset-default`](https://github.com/WordPress/gutenberg/tree/HEAD/packages/babel-preset-default#polyfill) in your code._

## Examples

### Data store

Preferences are persisted values of any kind.

Set the default preferences for any features on initialization by dispatching an action:

```js
import { dispatch } from '@wordpress/data';
import { store as preferencesStore } from '@wordpress/preferences';

function initialize() {
	// ...

	dispatch( preferencesStore ).setDefaults(
		'namespace/editor-or-plugin-name',
		{
			myBooleanFeature: true,
		}
	);

	// ...
}
```

Or the `get` selector to get a preference value, and the `set` action to update a preference to any value:

```js
wp.data
	.select( 'core/preferences' )
	.get( 'namespace/editor-or-plugin-name', 'myPreferenceName' ); // 1
wp.data
	.dispatch( 'core/preferences' )
	.set( 'namespace/editor-or-plugin-name', 'myPreferenceName', 2 );
wp.data
	.select( 'core/preferences' )
	.get( 'namespace/editor-or-plugin-name', 'myPreferenceName' ); // 2
```

Use the `toggle` action to flip a boolean preference between `true` and `false`:

```js
wp.data
	.select( 'core/preferences' )
	.get( 'namespace/editor-or-plugin-name', 'myPreferenceName' ); // true
wp.data
	.dispatch( 'core/preferences' )
	.toggle( 'namespace/editor-or-plugin-name', 'myPreferenceName' );
wp.data
	.select( 'core/preferences' )
	.get( 'namespace/editor-or-plugin-name', 'myPreferenceName' ); // false
```

### Components

The `PreferenceToggleMenuItem` components can be used with a `DropdownMenu` to implement a menu for changing preferences.

Also see the `MoreMenuDropdown` component from the `@wordpress/interface` package for implementing a more menu.

```jsx
function MyEditorMenu() {
	return (
		<MoreMenuDropdown>
			{ () => (
				<MenuGroup label={ __( 'Features' ) }>
					<PreferenceToggleMenuItem
						scope="namespace/editor-or-plugin-name"
						name="myPreferenceName"
						label={ __( 'My feature' ) }
						info={ __( 'A really awesome feature' ) }
						messageActivated={ __( 'My feature activated' ) }
						messageDeactivated={ __( 'My feature deactivated' ) }
					/>
				</MenuGroup>
			) }
		</MoreMenuDropdown>
	);
}
```

## API Reference

### Actions

The following set of dispatching action creators are available on the object returned by `wp.data.dispatch( 'core/preferences' )`:

<!-- START TOKEN(Autogenerated actions|src/store/actions.js) -->

#### set

Returns an action object used in signalling that a preference should be set
to a value

_Parameters_

-   _scope_ `string`: The preference scope (e.g. core/edit-post).
-   _name_ `string`: The preference name.
-   _value_ `*`: The value to set.

_Returns_

-   `Object`: Action object.

#### setDefaults

Returns an action object used in signalling that preference defaults should
be set.

_Parameters_

-   _scope_ `string`: The preference scope (e.g. core/edit-post).
-   _defaults_ `Object<string, *>`: A key/value map of preference names to values.

_Returns_

-   `Object`: Action object.

#### toggle

Returns an action object used in signalling that a preference should be
toggled.

_Parameters_

-   _scope_ `string`: The preference scope (e.g. core/edit-post).
-   _name_ `string`: The preference name.

<!-- END TOKEN(Autogenerated actions|src/store/actions.js) -->

### Selectors

The following selectors are available on the object returned by `wp.data.select( 'core/preferences' )`:

<!-- START TOKEN(Autogenerated selectors|src/store/selectors.js) -->

#### get

Returns a boolean indicating whether a prefer is active for a particular
scope.

_Parameters_

-   _state_ `Object`: The store state.
-   _scope_ `string`: The scope of the feature (e.g. core/edit-post).
-   _name_ `string`: The name of the feature.

_Returns_

-   `*`: Is the feature enabled?

<!-- END TOKEN(Autogenerated selectors|src/store/selectors.js) -->

## Contributing to this package

This is an individual package that's part of the Gutenberg project. The project is organized as a monorepo. It's made up of multiple self-contained software packages, each with a specific purpose. The packages in this monorepo are published to [npm](https://www.npmjs.com/) and used by [WordPress](https://make.wordpress.org/core/) as well as other software projects.

To find out more about contributing to this package or Gutenberg as a whole, please read the project's main [contributor guide](https://github.com/WordPress/gutenberg/tree/HEAD/CONTRIBUTING.md).

<br /><br /><p align="center"><img src="https://s.w.org/style/images/codeispoetry.png?1" alt="Code is Poetry." /></p>
